<?xml version = "1.0" encoding = "UTF-8"?>

<!--
    MDDL - Market Data Definition Language
    Version 2.0-beta 15 September 2002
    Copyright (c)2001-2002, The Software & Information Industry Association.
    See http://www.mddl.org/ or http://www.fisd.net/mddl/
-->

<xsl:transform xmlns:xsl = "http://www.w3.org/1999/XSL/Transform" version = "1.0">
	<xsl:output method = "text"/>
	<xsl:param name = "mddl-error" select = "'!!!! MDDL Schema Error: '"/>
	<xsl:param name = "mddl-warning" select = "'** MDDL Schema Warning: '"/>
	<xsl:param name = "mddl-info" select = "'== MDDL Schema Info: '"/>
	<xsl:variable name = "space" select = "' '"/>
	<xsl:variable name = "sep" select = "'; '"/>
	<xsl:key name="elementsByName" match="element[@name]" use="@name"/>
	<xsl:key name="domainsByName" match="domain[@name]" use="@name"/>
	<xsl:key name="classesByName" match="class[@name]" use="@name"/>
	<xsl:key name="subclassesByName" match="subclass[@name]" use="@name"/>
	<xsl:key name="propertiesByName" match="property[@name]" use="@name"/>
	<xsl:key name="groupsByName" match="group[@name]" use="@name"/>
	<xsl:key name="anyByRef" match="*[@ref]" use="concat(name(), '[@', @ref, ']')"/>
	<xsl:template match = "/">
		<!--Validate all elements.-->
		<xsl:apply-templates select = "*" mode = "validate"/>
		<xsl:message>
			<xsl:text>MDDL Schema validation completed.</xsl:text>
		</xsl:message>
	</xsl:template>
	<xsl:template match = "*" mode = "validate">
		<!--Default is just to validate children.-->
		<xsl:apply-templates select = "*" mode = "validate"/>
	</xsl:template>
	<xsl:template match = "element" mode = "validate">
		<!--Validates MDDL element definitions.-->
		
		<!--1. An "element" with a "name" must have either a type or element/domain/property/group children, but not both.-->
		<xsl:if test = "@name and not(@type) and count(*) = 0">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>an element definition without a type must have element/domain/property/group children.  Use type='none' for empty elements</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<xsl:if test = "@name and @type and (count(*) > 0)">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>an element definition with a type must not have element/domain/property/group children</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--2. An "element" with a "ref" cannot have any other attributes, nor any children.-->
		<xsl:if test = "@ref and (count(@*) > count(@ref | @use))">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>an element reference cannot have any extra attributes except 'use'</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<xsl:if test = "@ref and (count(*) > 0)">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>an element reference cannot have any children</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--3. An "element" with a "ref" cannot be a direct child of "mddl-schema".-->
		<xsl:if test = "@ref and parent::mddl-schema">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>an element reference cannot be a direct child of 'mddl-schema'</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--4. An "element" with a "use" cannot be a direct child of "mddl-schema".-->
		<xsl:if test = "@use and parent::mddl-schema">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a top-level element definition cannot have a "use" attribute</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--5. A "group" in an "element" may not contain classes nor subclasses, neither directly nor indirectly.-->
		<xsl:call-template name = "check-child">
			<xsl:with-param name = "elementType" select = "'class'"/>
		</xsl:call-template>
		<xsl:call-template name = "check-child">
			<xsl:with-param name = "elementType" select = "'subclass'"/>
		</xsl:call-template>
		<!--6. Although elements can have direct or indirect "property" children, this is strongly discouraged.-->
		<xsl:call-template name = "check-child">
			<xsl:with-param name = "elementType" select = "'property'"/>
			<xsl:with-param name = "errorLevel" select = "'warning'"/>
		</xsl:call-template>
		<!--7. Element names must be in "lowerCamelCase".-->
		<xsl:if test = "@name">
			<xsl:call-template name = "check-name">
				<xsl:with-param name = "name" select = "@name"/>
			</xsl:call-template>
		</xsl:if>
		<!--8. Element names must be unique.-->
		<xsl:if test = "@name">
			<xsl:variable name = "name" select = "@name"/>
			<xsl:if test = "count(key('elementsByName', $name)) != 1">
				<xsl:call-template name = "error-message">
					<xsl:with-param name = "message">
						<xsl:text>a named element can only be defined once</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
		</xsl:if>
		<!--9. Element references must have matching named element definitions.-->
		<xsl:if test = "@ref">
			<xsl:variable name = "ref" select = "@ref"/>
			<xsl:if test = "count(key('elementsByName', $ref)) = 0">
				<xsl:call-template name = "error-message">
					<xsl:with-param name = "message">
						<xsl:text>the referenced element has not been defined</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
		</xsl:if>
		<xsl:apply-templates select = "*" mode = "validate"/>
	</xsl:template>
	<xsl:template match = "domain" mode = "validate">
		<!--Validates MDDL domain definitions.-->
		
		<!--1. A domain must have a "name" or a "ref", but not both.-->
		<xsl:if test = "not(@name or @ref)">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a domain must have a "name" or a "ref" attribute</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<xsl:if test = "@name and @ref">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a domain must not have both a "name" and a "ref" attribute</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--2. A domain with a "ref" cannot have any children.-->
		<xsl:if test = "@ref and (count(*) > 0)">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a domain reference cannot have any children</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--3. A "domain" with a "ref" cannot be a direct child of "mddl-schema".-->
		<xsl:if test = "@ref and parent::mddl-schema">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a domain reference cannot be a direct child of 'mddl-schema'</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--4. A "group" in a "domain" may not contain domains nor elements, neither directly nor indirectly.-->
		<xsl:call-template name = "check-child">
			<xsl:with-param name = "elementType" select = "'element'"/>
		</xsl:call-template>
		<xsl:call-template name = "check-child">
			<xsl:with-param name = "elementType" select = "'domain'"/>
		</xsl:call-template>
		<!--5. Domain names should not include the "Domain" suffix.-->
		<xsl:if test = "@name and (string-length(@name) >= string-length('Domain'))">
			<xsl:variable name = "suffixLength" select = "string-length('Domain')"/>
			<xsl:variable name = "suffix" select = "substring(@name, string-length(@name) - $suffixLength + 1, $suffixLength)"/>
			<xsl:if test = "($suffix = 'Domain') or ($suffix = 'domain')">
				<xsl:call-template name = "warning-message">
					<xsl:with-param name = "message">
						<xsl:text>domain names should not include the 'Domain' suffix, as this is added automatically</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
		</xsl:if>
		<!--6. Domain names must be in "lowerCamelCase".-->
		<xsl:if test = "@name">
			<xsl:call-template name = "check-name">
				<xsl:with-param name = "name" select = "@name"/>
			</xsl:call-template>
		</xsl:if>
		<!--7. Domain names must be unique.-->
		<xsl:if test = "@name">
			<xsl:variable name = "name" select = "@name"/>
			<xsl:if test = "count(key('domainsByName', $name)) != 1">
				<xsl:call-template name = "error-message">
					<xsl:with-param name = "message">
						<xsl:text>a named domain can only be defined once</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
		</xsl:if>
		<!--8. Domain references must have matching named domain definitions.-->
		<xsl:if test = "@ref">
			<xsl:variable name = "ref" select = "@ref"/>
			<xsl:if test = "count(key('domainsByName', $ref)) = 0">
				<xsl:call-template name = "error-message">
					<xsl:with-param name = "message">
						<xsl:text>the referenced domain has not been defined</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
		</xsl:if>
		<xsl:apply-templates select = "*" mode = "validate"/>
	</xsl:template>
	<xsl:template match = "class" mode = "validate">
		<!--Validates MDDL class definitions.-->
		
		<!--1. A class must have a "name" or a "ref", but not both.-->
		<xsl:if test = "not(@name or @ref)">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a class must have a "name" or a "ref" attribute</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<xsl:if test = "@name and @ref">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a class must not have both a "name" and a "ref" attribute</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--2. A class with a "ref" cannot have any children.-->
		<xsl:if test = "@ref and (count(*) > 0)">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a class reference cannot have any children</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--3. A "class" with a "ref" cannot be a direct child of "mddl-schema".-->
		<xsl:if test = "@ref and parent::mddl-schema">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a class reference cannot be a direct child of 'mddl-schema'</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--4. A "group" in a "class" may not contain classes nor domains nor elements, neither directly nor indirectly.-->
		<xsl:call-template name = "check-child">
			<xsl:with-param name = "elementType" select = "'element'"/>
		</xsl:call-template>
		<xsl:call-template name = "check-child">
			<xsl:with-param name = "elementType" select = "'domain'"/>
		</xsl:call-template>
		<xsl:call-template name = "check-child">
			<xsl:with-param name = "elementType" select = "'class'"/>
		</xsl:call-template>
		<!--5. Class names should not include the "Class" suffix.-->
		<xsl:if test = "@name and (string-length(@name) >= string-length('Class'))">
			<xsl:variable name = "suffixLength" select = "string-length('Class')"/>
			<xsl:variable name = "suffix" select = "substring(@name, string-length(@name) - $suffixLength + 1, $suffixLength)"/>
			<xsl:if test = "($suffix = 'Class') or ($suffix = 'class')">
				<xsl:call-template name = "warning-message">
					<xsl:with-param name = "message">
						<xsl:text>class names should not include the 'Class' suffix, as this is added automatically</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
		</xsl:if>
		<!--6. Class names must be in "lowerCamelCase".-->
		<xsl:if test = "@name">
			<xsl:call-template name = "check-name">
				<xsl:with-param name = "name" select = "@name"/>
			</xsl:call-template>
		</xsl:if>
		<!--7. Class names must be unique.-->
		<xsl:if test = "@name">
			<xsl:variable name = "name" select = "@name"/>
			<xsl:if test = "count(key('classesByName', $name)) != 1">
				<xsl:call-template name = "error-message">
					<xsl:with-param name = "message">
						<xsl:text>a named class can only be defined once</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
		</xsl:if>
		<!--8. Class references must have matching named class definitions.-->
		<xsl:if test = "@ref">
			<xsl:variable name = "ref" select = "@ref"/>
			<xsl:if test = "count(key('classesByName', $ref)) = 0">
				<xsl:call-template name = "error-message">
					<xsl:with-param name = "message">
						<xsl:text>the referenced class has not been defined</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
		</xsl:if>
		<xsl:apply-templates select = "*" mode = "validate"/>
	</xsl:template>
	<xsl:template match = "subclass" mode = "validate">
		<!--Validates MDDL subclass definitions.-->
		
		<!--1. A subclass must have a "name" or a "ref", but not both.-->
		<xsl:if test = "not(@name or @ref)">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a subclass must have a "name" or a "ref" attribute</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<xsl:if test = "@name and @ref">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a subclass must not have both a "name" and a "ref" attribute</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--2. A subclass with a "ref" cannot have any children.-->
		<xsl:if test = "@ref and (count(*) > 0)">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a subclass reference cannot have any children</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--3. A "subclass" with a "ref" cannot be a direct child of "mddl-schema".-->
		<xsl:if test = "@ref and parent::mddl-schema">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a subclass reference cannot be a direct child of 'mddl-schema'</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--4. A "group" in a "subclass" may not contain subclasses nor classes nor domains nor elements, neither directly nor indirectly.-->
		<xsl:call-template name = "check-child">
			<xsl:with-param name = "elementType" select = "'element'"/>
		</xsl:call-template>
		<xsl:call-template name = "check-child">
			<xsl:with-param name = "elementType" select = "'domain'"/>
		</xsl:call-template>
		<xsl:call-template name = "check-child">
			<xsl:with-param name = "elementType" select = "'class'"/>
		</xsl:call-template>
		<xsl:call-template name = "check-child">
			<xsl:with-param name = "elementType" select = "'subclass'"/>
		</xsl:call-template>
		<!--5. Subclass names should not include the "Subclass" suffix.-->
		<xsl:if test = "@name and (string-length(@name) >= string-length('Subclass'))">
			<xsl:variable name = "suffixLength" select = "string-length('Subclass')"/>
			<xsl:variable name = "suffix" select = "substring(@name, string-length(@name) - $suffixLength + 1, $suffixLength)"/>
			<xsl:if test = "($suffix = 'Subclass') or ($suffix = 'subclass')">
				<xsl:call-template name = "warning-message">
					<xsl:with-param name = "message">
						<xsl:text>subclass names should not include the 'Subclass' suffix, as this is added automatically</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
		</xsl:if>
		<!--6. Subclass names must be in "lowerCamelCase".-->
		<xsl:if test = "@name">
			<xsl:call-template name = "check-name">
				<xsl:with-param name = "name" select = "@name"/>
			</xsl:call-template>
		</xsl:if>
		<!--7. Subclass names must be unique.-->
		<xsl:if test = "@name">
			<xsl:variable name = "name" select = "@name"/>
			<xsl:if test = "count(key('subclassesByName', $name)) != 1">
				<xsl:call-template name = "error-message">
					<xsl:with-param name = "message">
						<xsl:text>a named subclass can only be defined once</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
		</xsl:if>
		<!--8. Subclass references must have matching named subclass definitions.-->
		<xsl:if test = "@ref">
			<xsl:variable name = "ref" select = "@ref"/>
			<xsl:if test = "count(key('subclassesByName', $ref)) = 0">
				<xsl:call-template name = "error-message">
					<xsl:with-param name = "message">
						<xsl:text>the referenced subclass has not been defined</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
		</xsl:if>
		<xsl:apply-templates select = "*" mode = "validate"/>
	</xsl:template>
	<xsl:template match = "property" mode = "validate">
		<!--Validates MDDL property definitions.-->
		
		<!--1. A property must have a "name" or a "ref", but not both.-->
		<xsl:if test = "not(@name or @ref)">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a property must have a "name" or a "ref" attribute</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<xsl:if test = "@name and @ref">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a property must not have both a "name" and a "ref" attribute</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--2. A property with a "name" must have a "type".-->
		<xsl:if test = "@name and not(@type)">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a property with a "name" attribute must have a "type" attribute</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--3. A property with a "ref" cannot have any children.-->
		<xsl:if test = "@ref and (count(*) > 0)">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a property reference cannot have any children</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--4. A property whose "type" is "none" cannot have any children (except attribute definitions).-->
		<xsl:if test = "@type and (@type = 'none') and (count(*) > count(attribute))">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a property whose type is "none" cannot have any children</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--5. A property cannot have a "default-scheme" URI if the type is "properties" or "none".-->
		<xsl:if test = "@default-scheme and (@type = 'properties' or @type = 'none')">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a property whose type is "properties" or "none" cannot have a "default-scheme"</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--6. A "subclass" with a "ref" cannot be a direct child of "mddl-schema".-->
		<xsl:if test = "@ref and parent::mddl-schema">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a property reference cannot be a direct child of 'mddl-schema'</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--7. A "group" in a "property" may not contain subclasses nor classes nor domains nor elements, neither directly nor indirectly.-->
		<xsl:call-template name = "check-child">
			<xsl:with-param name = "elementType" select = "'element'"/>
		</xsl:call-template>
		<xsl:call-template name = "check-child">
			<xsl:with-param name = "elementType" select = "'domain'"/>
		</xsl:call-template>
		<xsl:call-template name = "check-child">
			<xsl:with-param name = "elementType" select = "'class'"/>
		</xsl:call-template>
		<xsl:call-template name = "check-child">
			<xsl:with-param name = "elementType" select = "'subclass'"/>
		</xsl:call-template>
		<!--8. A "property" may not contain itself as a child, neither directly nor indirectly.  That is to say, recursive property definitions are not allowed.-->
		<xsl:if test = "@name">
			<xsl:call-template name = "check-resursive-property">
				<xsl:with-param name = "targetName" select = "@name"/>
			</xsl:call-template>
		</xsl:if>
		<!--9. Property names should not include the "Property" suffix.-->
		<xsl:if test = "@name and (string-length(@name) >= string-length('Property'))">
			<xsl:variable name = "suffixLength" select = "string-length('Property')"/>
			<xsl:variable name = "suffix" select = "substring(@name, string-length(@name) - $suffixLength + 1, $suffixLength)"/>
			<xsl:if test = "($suffix = 'Property') or ($suffix = 'property')">
				<xsl:call-template name = "warning-message">
					<xsl:with-param name = "message">
						<xsl:text>property names should not include the 'Property' suffix</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
		</xsl:if>
		<!--10. Property names must be in "lowerCamelCase".-->
		<xsl:if test = "@name">
			<xsl:call-template name = "check-name">
				<xsl:with-param name = "name" select = "@name"/>
			</xsl:call-template>
		</xsl:if>
		<!--11. Property names must be unique.-->
		<xsl:if test = "@name">
			<xsl:variable name = "name" select = "@name"/>
			<xsl:if test = "count(key('propertiesByName', $name)) != 1">
				<xsl:call-template name = "error-message">
					<xsl:with-param name = "message">
						<xsl:text>a named property can only be defined once</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
		</xsl:if>
		<!--12. Property references must have matching named property definitions.-->
		<xsl:if test = "@ref">
			<xsl:variable name = "ref" select = "@ref"/>
			<xsl:if test = "count(key('propertiesByName', $ref)) = 0">
				<xsl:call-template name = "error-message">
					<xsl:with-param name = "message">
						<xsl:text>the referenced property has not been defined</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
		</xsl:if>
		<xsl:apply-templates select = "*" mode = "validate"/>
	</xsl:template>
	<xsl:template match = "group" mode = "validate">
		<!--Validates MDDL group definitions.-->
		
		<!--1. A group cannot have both a "name" and a "ref".-->
		<xsl:if test = "@name and @ref">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a group must not have both a "name" and a "ref" attribute</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--2. A group with a "ref" cannot have any children, nor a "method".-->
		<xsl:if test = "@ref and (count(*) > 0)">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a group reference cannot have any children</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<xsl:if test = "@ref and @method">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a group reference cannot have a "method" attribute</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--3. A "group" without a "name" cannot be a direct child of "mddl-schema".-->
		<xsl:if test = "@ref and parent::mddl-schema">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>a group reference cannot be a direct child of 'mddl-schema'</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<xsl:if test = "not(@name) and parent::mddl-schema">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>an unnamed group cannot be a direct child of 'mddl-schema'</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--4. Group names should not include a ".model" or ".group" suffix.-->
		<xsl:if test = "@name and (string-length(@name) >= string-length('.model'))">
			<xsl:variable name = "suffixLength" select = "string-length('.model')"/>
			<xsl:variable name = "suffix" select = "substring(@name, string-length(@name) - $suffixLength + 1, $suffixLength)"/>
			<xsl:if test = "($suffix = '.model') or ($suffix = '.Model')">
				<xsl:call-template name = "warning-message">
					<xsl:with-param name = "message">
						<xsl:text>domain names should not include the '.model' suffix, as this is added automatically</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
		</xsl:if>
		<xsl:if test = "@name and (string-length(@name) >= string-length('.group'))">
			<xsl:variable name = "suffixLength" select = "string-length('.group')"/>
			<xsl:variable name = "suffix" select = "substring(@name, string-length(@name) - $suffixLength + 1, $suffixLength)"/>
			<xsl:if test = "($suffix = '.group') or ($suffix = '.Group')">
				<xsl:call-template name = "warning-message">
					<xsl:with-param name = "message">
						<xsl:text>domain names should not include the '.group' suffix, as the ".model" suffix is added automatically</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
		</xsl:if>
		<!--5. Group names must be unique, and may not be the same as the name of any element/domain/class/subclass/property.-->
		<xsl:if test = "@name">
			<xsl:variable name = "name" select = "@name"/>
			<xsl:if test = "count(key('groupsByName', $name)) != 1">
				<xsl:call-template name = "error-message">
					<xsl:with-param name = "message">
						<xsl:text>a named group can only be defined once</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
			<xsl:if test = "count(key('elementsByName', $name)) > 0">
				<xsl:call-template name = "error-message">
					<xsl:with-param name = "message">
						<xsl:text>a group may not have the same name as an element</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
			<xsl:if test = "count(key('domainsByName', $name)) > 0">
				<xsl:call-template name = "error-message">
					<xsl:with-param name = "message">
						<xsl:text>a group may not have the same name as a domain</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
			<xsl:if test = "count(key('classesByName', $name)) > 0">
				<xsl:call-template name = "error-message">
					<xsl:with-param name = "message">
						<xsl:text>a group may not have the same name as a class</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
			<xsl:if test = "count(key('subclassesByName', $name)) > 0">
				<xsl:call-template name = "error-message">
					<xsl:with-param name = "message">
						<xsl:text>a group may not have the same name as a subclass</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
			<xsl:if test = "count(key('propertiesByName', $name)) > 0">
				<xsl:call-template name = "error-message">
					<xsl:with-param name = "message">
						<xsl:text>a group may not have the same name as a property</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
		</xsl:if>
		<!--7. Group references must have matching named group definitions.-->
		<xsl:if test = "@ref">
			<xsl:variable name = "ref" select = "@ref"/>
			<xsl:if test = "count(key('groupsByName', $ref)) = 0">
				<xsl:call-template name = "error-message">
					<xsl:with-param name = "message">
						<xsl:text>the referenced group has not been defined</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:if>
		</xsl:if>
		<xsl:apply-templates select = "*" mode = "validate"/>
	</xsl:template>
	<xsl:template match = "attribute" mode = "validate">
		<!--Validates MDDL attribute definitions.-->
		
		<!--1. An attribute can only have a value if its use is not "required" (for "optional", the value is treated as the default value).-->
		<xsl:if test = "@value and (@use = 'required')">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>an attribute can only have a value if its use is not "required"</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--2. An attribute which is "fixed" must have a "value".-->
		<xsl:if test = "(@use = 'fixed') and not(@value)">
			<xsl:call-template name = "error-message">
				<xsl:with-param name = "message">
					<xsl:text>an attribute whose use is "fixed" must have a "value" attribute</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<!--3. Although properties can have attributes, this is strongly discouraged.-->
		<xsl:if test = "parent::property">
			<xsl:call-template name = "warning-message">
				<xsl:with-param name = "message">
					<xsl:text>attributes for properties are discouraged.  Use a child property instead if possible</xsl:text>
				</xsl:with-param>
			</xsl:call-template>
		</xsl:if>
		<xsl:apply-templates select = "*" mode = "validate"/>
	</xsl:template>
	<xsl:template name = "error-message">
		<xsl:param name = "message"/>
		<xsl:param name = "trace"/>
		<xsl:message terminate = "yes">
			<xsl:value-of select = "$mddl-error"/>
			<xsl:value-of select = "$message"/>
			<xsl:text>:</xsl:text>
			<xsl:value-of select = "$space"/>
			<xsl:choose>
				<xsl:when test = "$trace">
					<xsl:value-of select = "$trace"/>
				</xsl:when>
				<xsl:otherwise>
					<xsl:call-template name = "write-location"/>
				</xsl:otherwise>
			</xsl:choose>
		</xsl:message>
	</xsl:template>
	<xsl:template name = "warning-message">
		<xsl:param name = "message"/>
		<xsl:param name = "trace"/>
		<xsl:message terminate = "no">
			<xsl:value-of select = "$mddl-warning"/>
			<xsl:value-of select = "$message"/>
			<xsl:text>:</xsl:text>
			<xsl:value-of select = "$space"/>
			<xsl:choose>
				<xsl:when test = "$trace">
					<xsl:value-of select = "$trace"/>
				</xsl:when>
				<xsl:otherwise>
					<xsl:call-template name = "write-location"/>
				</xsl:otherwise>
			</xsl:choose>
		</xsl:message>
	</xsl:template>
	<xsl:template name = "info-message">
		<xsl:param name = "message"/>
		<xsl:param name = "trace"/>
		<xsl:message terminate = "no">
			<xsl:value-of select = "$mddl-info"/>
			<xsl:value-of select = "$message"/>
			<xsl:text>:</xsl:text>
			<xsl:value-of select = "$space"/>
			<xsl:choose>
				<xsl:when test = "$trace">
					<xsl:value-of select = "$trace"/>
				</xsl:when>
				<xsl:otherwise>
					<xsl:call-template name = "write-location"/>
				</xsl:otherwise>
			</xsl:choose>
		</xsl:message>
	</xsl:template>
	<xsl:template name = "write-location">
		<xsl:text>&lt;</xsl:text>
		<xsl:value-of select = "name()"/>
		<xsl:for-each select = "@*">
			<xsl:value-of select = "$space"/>
			<xsl:value-of select = "name()"/>
			<xsl:text>="</xsl:text>
			<xsl:value-of select = "."/>
			<xsl:text>"</xsl:text>
		</xsl:for-each>
		<xsl:if test = "count(node()) = 0">
			<xsl:text>/</xsl:text>
		</xsl:if>
		<xsl:text>></xsl:text>
	</xsl:template>
	<xsl:template name = "check-name">
		<xsl:param name = "name"/>
		<xsl:param name = "errorLevel" select = "'error'"/>
		<xsl:variable name = "lowerChars">
			<xsl:text>abcdefghijklmnopqrstuvwxyz</xsl:text>
		</xsl:variable>
		<xsl:variable name = "allChars">
			<xsl:value-of select = "$lowerChars"/>
			<xsl:text>ABCDEFGHIJKLMNOPQRSTUVWXYZ</xsl:text>
		</xsl:variable>
		<!--Names of elements/domains/classes/subclasses/properties must be lowerCamelCase.-->
		
		<!--They must also be constructed from the letters [A-Za-z].-->
		
		<!--This template does not do a full check for lowerCamelCase, but a plausibility check.-->
		<xsl:variable name = "isLowerFirst">
			<xsl:call-template name = "check-valid-string">
				<xsl:with-param name = "string" select = "substring($name, 1, 1)"/>
				<xsl:with-param name = "validChars" select = "$lowerChars"/>
			</xsl:call-template>
		</xsl:variable>
		<xsl:if test = "normalize-space($isLowerFirst) != ''">
			<xsl:choose>
				<xsl:when test = "$errorLevel = 'warning'">
					<xsl:call-template name = "warning-message">
						<xsl:with-param name = "message">
							<xsl:text>the first character of the name should be lower case [a-z]</xsl:text>
						</xsl:with-param>
					</xsl:call-template>
				</xsl:when>
				<xsl:otherwise>
					<xsl:call-template name = "error-message">
						<xsl:with-param name = "message">
							<xsl:text>the first character of the name must be lower case [a-z]</xsl:text>
						</xsl:with-param>
					</xsl:call-template>
				</xsl:otherwise>
			</xsl:choose>
		</xsl:if>
		<xsl:variable name = "isAlpha">
			<xsl:call-template name = "check-valid-string">
				<xsl:with-param name = "string" select = "substring($name, 2)"/>
				<xsl:with-param name = "validChars" select = "$allChars"/>
			</xsl:call-template>
		</xsl:variable>
		<xsl:if test = "normalize-space($isAlpha) != ''">
			<xsl:choose>
				<xsl:when test = "$errorLevel = 'warning'">
					<xsl:call-template name = "warning-message">
						<xsl:with-param name = "message">
							<xsl:text>all characters in the name should be alphabetic [A-Za-z]</xsl:text>
						</xsl:with-param>
					</xsl:call-template>
				</xsl:when>
				<xsl:otherwise>
					<xsl:call-template name = "error-message">
						<xsl:with-param name = "message">
							<xsl:text>all characters in the name must be alphabetic [A-Za-z]</xsl:text>
						</xsl:with-param>
					</xsl:call-template>
				</xsl:otherwise>
			</xsl:choose>
		</xsl:if>
	</xsl:template>
	<xsl:template name = "check-valid-string">
		<xsl:param name = "string"/>
		<xsl:param name = "validChars"/>
		<xsl:if test = "$string != ''">
			<xsl:if test = "not(contains($validChars, substring($string, 1, 1)))">
				<xsl:value-of select = "substring($string, 1, 1)"/>
			</xsl:if>
			<xsl:call-template name = "check-valid-string">
				<xsl:with-param name = "string" select = "substring($string, 2)"/>
				<xsl:with-param name = "validChars" select = "$validChars"/>
			</xsl:call-template>
		</xsl:if>
	</xsl:template>
	<xsl:template name = "check-child">
		<!--Check for a child of the given MDDL-Schema element type (element/domain/class/subclass/property).-->
		<xsl:param name = "elementType"/>
		<xsl:param name = "trace"/>
		<xsl:param name = "errorLevel" select = "'error'"/>
		<xsl:variable name = "parentTrace">
			<xsl:if test = "$trace">
				<xsl:value-of select = "$trace"/>
				<xsl:value-of select = "$sep"/>
			</xsl:if>
			<xsl:call-template name = "write-location"/>
		</xsl:variable>
		<xsl:for-each select = "*[name() = $elementType]">
			<xsl:choose>
				<xsl:when test = "$errorLevel = 'warning'">
					<xsl:call-template name = "warning-message">
						<xsl:with-param name = "message">
							<xsl:text>MDDL Schema: use of a child '</xsl:text>
							<xsl:value-of select = "name()"/>
							<xsl:text>' is discouraged here</xsl:text>
						</xsl:with-param>
						<xsl:with-param name = "trace">
							<xsl:value-of select = "$parentTrace"/>
							<xsl:value-of select = "$sep"/>
							<xsl:call-template name = "write-location"/>
						</xsl:with-param>
					</xsl:call-template>
				</xsl:when>
				<xsl:otherwise>
					<xsl:call-template name = "error-message">
						<xsl:with-param name = "message">
							<xsl:text>invalid MDDL Schema: child '</xsl:text>
							<xsl:value-of select = "name()"/>
							<xsl:text>' not allowed here</xsl:text>
						</xsl:with-param>
						<xsl:with-param name = "trace">
							<xsl:value-of select = "$parentTrace"/>
							<xsl:value-of select = "$sep"/>
							<xsl:call-template name = "write-location"/>
						</xsl:with-param>
					</xsl:call-template>
				</xsl:otherwise>
			</xsl:choose>
		</xsl:for-each>
		<xsl:for-each select = "group[not(@ref)]">
			<xsl:call-template name = "check-child">
				<xsl:with-param name = "elementType" select = "$elementType"/>
				<xsl:with-param name = "trace" select = "$parentTrace"/>
				<xsl:with-param name = "errorLevel" select = "$errorLevel"/>
			</xsl:call-template>
		</xsl:for-each>
		<xsl:for-each select = "group[@ref]">
			<xsl:variable name = "ref" select = "@ref"/>
			<xsl:for-each select = "key('groupsByName', $ref)">
				<xsl:call-template name = "check-child">
					<xsl:with-param name = "elementType" select = "$elementType"/>
					<xsl:with-param name = "trace" select = "$parentTrace"/>
					<xsl:with-param name = "errorLevel" select = "$errorLevel"/>
				</xsl:call-template>
			</xsl:for-each>
		</xsl:for-each>
	</xsl:template>
	<!--Check whether the given property calls itself recursively.-->
	<xsl:template name = "check-resursive-property">
		<xsl:param name = "targetName"/>
		<xsl:param name = "currentName" select = "$targetName"/>
		<xsl:param name = "path" select = "concat('[', $targetName, ']')"/>
		<xsl:variable name = "elementName" select = "name()"/>
		<xsl:for-each select = "key('anyByRef', concat($elementName, '[@', $currentName, ']'))">
			<xsl:for-each select = "ancestor::property[contains($path, concat('[', @name, ']'))]">
				<xsl:call-template name = "error-message">
					<xsl:with-param name = "message">
						<xsl:text>invalid MDDL Schema: property '</xsl:text>
						<xsl:value-of select = "@name"/>
						<xsl:text>' contains itself as a child recursively (</xsl:text>
						<xsl:value-of select = "translate(concat(@name, '/', $path), ' []', '/')"/>
						<xsl:text>)</xsl:text>
					</xsl:with-param>
				</xsl:call-template>
			</xsl:for-each>
			<xsl:for-each select = "ancestor::*[@name][1][not(self::element)]">
				<xsl:call-template name = "check-resursive-property">
					<xsl:with-param name = "targetName" select = "$targetName"/>
					<xsl:with-param name = "currentName" select = "@name"/>
					<xsl:with-param name = "path" select = "concat('[', @name, '] ', $path)"/>
				</xsl:call-template>
			</xsl:for-each>
		</xsl:for-each>
	</xsl:template>
</xsl:transform>